from typing import cast

immutable_builtin_types = (
    int,
    float,
    complex,
    str,
    bool,
    cast(type, frozenset),
    bytes,
)

immutable_but_containable_builtin_types = (cast(type, tuple),)

mutable_builtin_types = (
    cast(type, list),
    bytearray,
    memoryview,
    cast(type, set),
    cast(type, dict),
)

supported_builtin_types = (
    *immutable_builtin_types,
    *immutable_but_containable_builtin_types,
    *mutable_builtin_types,
)



# pyright: reportMissingTypeStubs=false

from reactivity.computed import ComputedRef, computed, is_computed_ref
from reactivity.effect import ReactiveEffect, effect
from reactivity.patches import patch
from reactivity.reactive import (deep_to_raw, is_reactive, mark_raw, reactive, to_raw)
from reactivity.ref import Ref, deep_unref, is_ref, ref, unref
from reactivity.watch import watch, watch_effect

from .__version__ import __version__

watchEffect = watch_effect
toRaw = to_raw
deepToRaw = deep_to_raw
isReactive = is_reactive
markRaw = mark_raw
deepUnref = deep_unref
isRef = is_ref
isComputedRef = is_computed_ref

patch()

__all__ = [
    'effect', 'watch_effect', 'watchEffect', 'watch', 'reactive', 'ref', 'computed', 'is_reactive', 'is_ref', 'unref',
    'deep_unref', 'deepUnref', '__version__', 'to_raw', 'toRaw', 'deep_to_raw', 'deepToRaw', 'isReactive', 'isRef',
    'mark_raw', 'markRaw', 'is_computed_ref', 'isComputedRef', 'Ref', 'ComputedRef', 'ReactiveEffect'
]



import json
import types
from typing import Any, Dict, Optional, Tuple, cast

from reactivity.ref.definitions import Ref
from reactivity.ref.utils import is_ref

__json_encoder_map: Dict[type, type] = {}

__raw_json_dump = json.dump
__raw_json_dumps = json.dumps


class PyReactivityJSONEncoderMetaClass(type):

    def __new__(cls, name: str, bases: Tuple[type, ...], attrs: Dict[str, Any]) -> Optional[type]:

        proxy_cls = type.__new__(cls, name, bases, attrs)
        raw_default = getattr(proxy_cls, 'default')

        def default(self: object, o: object):
            return cast(Ref[Any], o).value if is_ref(o) else raw_default(self, o)

        setattr(proxy_cls, 'default', default)

        return proxy_cls


def __patch_json_dumps():

    def dumps(*args: Any, **kwargs: Any):
        if 'default' in kwargs:
            raw_default = kwargs['default']

            def default(o: object):
                return cast(Ref[Any], o).value if is_ref(o) else raw_default(o)

            kwargs['default'] = default

        raw_class = kwargs['cls'] if 'cls' in kwargs else json.JSONEncoder
        if raw_class not in __json_encoder_map:
            class_name = raw_class.__name__
            __json_encoder_map[raw_class] = types.new_class(
                class_name,
                (raw_class,),
                {'metaclass': PyReactivityJSONEncoderMetaClass},
            )
        kwargs['cls'] = __json_encoder_map[raw_class]

        return __raw_json_dumps(*args, **kwargs)

    json.dumps = dumps


def __patch_json_dump():

    def dump(*args: Any, **kwargs: Any):
        if 'default' in kwargs:
            raw_default = kwargs['default']

            def default(o: object):
                return cast(Ref[Any], o).value if is_ref(o) else raw_default(o)

            kwargs['default'] = default

        raw_class = kwargs['cls'] if 'cls' in kwargs else json.JSONEncoder
        if raw_class not in __json_encoder_map:
            class_name = raw_class.__name__
            __json_encoder_map[raw_class] = types.new_class(
                class_name,
                (raw_class,),
                {'metaclass': PyReactivityJSONEncoderMetaClass},
            )
        kwargs['cls'] = __json_encoder_map[raw_class]

        return __raw_json_dump(*args, **kwargs)

    json.dump = dump


def patch_json():
    __patch_json_dumps()
    __patch_json_dump()


def unpatch_json():
    json.dumps = __raw_json_dumps
    json.dump = __raw_json_dump


def default(o: object):
    if is_ref(o):
        return cast(Ref[Any], o).value
    raise TypeError(f'Object of type {o.__class__.__name__} is not JSON serializable')



# pyright: reportMissingTypeStubs=false

from typing import Any, Dict, Iterable, TypeVar, Union, cast, overload

from reactivity.flags import FLAG_OF_REF

from .definitions import Ref

T = TypeVar('T')


def is_ref(obj: object) -> bool:
    return hasattr(obj, FLAG_OF_REF)


@overload
def unref(obj: Ref[T]) -> T:
    ...


@overload
def unref(obj: T) -> T:
    ...


def unref(obj: Union[Ref[T], T]) -> T:
    if is_ref(obj):
        obj = cast(Ref[T], obj)
        return obj.value
    obj = cast(T, obj)
    return obj


@overload
def deep_unref(obj: Ref[T]) -> T:
    ...


@overload
def deep_unref(obj: T) -> T:
    ...


def deep_unref(obj: 'Union[Ref[T], T]') -> T:
    if is_ref(obj):
        obj = cast(Ref[T], obj)
        v = obj.value
    else:
        v = cast(T, obj)

    if isinstance(v, dict):
        v = cast(Dict[Any, Any], v)
        return cast(T, {k: deep_unref(v) for k, v in v.items()})
    elif isinstance(v, list):
        v = cast(Iterable[Any], v)
        return cast(T, [deep_unref(i) for i in v])
    elif isinstance(v, tuple):
        v = cast(Iterable[Any], v)
        return cast(T, tuple(deep_unref(i) for i in v))
    elif isinstance(v, set):
        v = cast(Iterable[Any], v)
        return cast(T, {deep_unref(i) for i in v})
    elif isinstance(v, frozenset):
        v = cast(Iterable[Any], v)
        return cast(T, frozenset(deep_unref(i) for i in v))
    return v



# pyright: reportMissingTypeStubs=false

from typing import Any, Callable, Dict, Generic, Set, TypeVar, Union, cast

from reactivity.effect import ReactiveEffect
from reactivity.flags import (FLAG_OF_COMPUTED_REF, FLAG_OF_READONLY, FLAG_OF_REF)
from reactivity.ref import track_ref_value, trigger_ref_value
from reactivity.reactive.utils import reactive_reversed_class_map

from .definitions import ComputedRef
from .utils import is_computed_ref  # pyright: ignore[reportUnusedImport]

T = TypeVar('T')


class ComputedRefImpl(Generic[T]):
    __value: Union[T, None]
    deps: Dict[Union[str, int], Set[ReactiveEffect[Any]]]
    effect: ReactiveEffect[T]
    _dirty: bool
    _cacheable: bool

    def __init__(self, getter: Callable[[], T]) -> None:
        self.__value = None
        self._dirty = True
        self._cacheable = True
        self.deps = {}

        def scheduler():
            if not self._dirty:
                self._dirty = True
                trigger_ref_value(self)

        self.effect = ReactiveEffect(getter, scheduler)
        self.effect.computed = self
        self.effect.active = self._cacheable
        setattr(self, FLAG_OF_COMPUTED_REF, True)
        setattr(self, FLAG_OF_REF, True)
        setattr(self, FLAG_OF_READONLY, True)

    @property
    def value(self) -> T:
        track_ref_value(self)
        if self._dirty:
            self._dirty = False
            self.__value = self.effect.run()
        return cast(T, self.__value)

    def __str__(self) -> str:
        t = type(self.__value)
        if t in reactive_reversed_class_map:
            t = reactive_reversed_class_map[t]
        return f'<ComputedRef[{t.__name__}] value={self.__value}>'


def computed(getter: Callable[[], T]) -> ComputedRef[T]:
    result = ComputedRefImpl(getter)
    return cast(ComputedRef[T], result)


__all__ = ['computed', 'is_computed_ref', 'ComputedRef']



# pyright: reportMissingTypeStubs=false

from typing import Any, Dict, Iterable, Set, TypeVar, Union, cast

from reactivity.effect.definations import ReactiveEffectDef
from reactivity.effect.utils import track_effects, trigger_effects
from reactivity.effect.vars import active_effect_stack
from reactivity.env import DEV
from reactivity.flags import FLAG_OF_REACTIVE, REACTIVITY_VALUE

T = TypeVar('T')

reactive_class_map: Dict[type, type] = {}
reactive_reversed_class_map: Dict[type, type] = {}
reactive_deps_map: 'Dict[int, Dict[Union[str, int], Set[ReactiveEffectDef[Any]]]]' = {}

__global_reactive_object_map: Dict[int, object] = {}
__global_original_object_map: Dict[int, object] = {}

__marked_raw_set: Set[int] = set()


def is_reactive(obj: object) -> bool:
    return hasattr(obj, FLAG_OF_REACTIVE)


def is_reactive_multable_sequence(obj: object) -> bool:
    return hasattr(obj, '__setitem__') and hasattr(obj, '__getitem__') and is_reactive(obj)


def __get_reactive_subscribers(obj: object, key: Union[str, int]) -> Set[ReactiveEffectDef[Any]]:
    obj_id = id(obj)
    if obj_id not in reactive_deps_map:
        reactive_deps_map[obj_id] = {}
    if key not in reactive_deps_map[obj_id]:
        reactive_deps_map[obj_id][key] = set()
    return reactive_deps_map[obj_id][key]


def track_reactive(obj: object, key: str) -> None:
    if not active_effect_stack:
        return
    deps = __get_reactive_subscribers(obj, key)
    track_effects(deps)


def track_reactive_value(obj: object) -> None:
    track_reactive(obj, REACTIVITY_VALUE)


def trigger_reactive(obj: object, key: str) -> None:
    subscribers = __get_reactive_subscribers(obj, key)
    trigger_effects(subscribers)


def trigger_reactive_value(obj: object) -> None:
    trigger_reactive(obj, REACTIVITY_VALUE)


def is_in_global_reactive_object_map(original: object) -> bool:
    return id(original) in __global_reactive_object_map


def is_in_global_original_object_map(observed: object) -> bool:
    return id(observed) in __global_original_object_map


def get_global_reactive_obj(original: T) -> T:
    return cast(T, __global_reactive_object_map[id(original)])


def to_raw(observed: T) -> T:
    observed_id = id(observed)
    if observed_id in __global_original_object_map:
        return cast(T, __global_original_object_map[observed_id])
    return observed


def deep_to_raw(observed: T) -> T:
    '''Recursively convert a reactive object to a plain python object.
    
    **Note:**
    
    The object returned is **not guaranteed** to be the original object.
    
    The only guarantee is that the object returned must be a plain python object.

    Args:
        observed (T): The object to be converted.

    Returns:
        T: The converted object, which must be a plain python object.
    '''
    if isinstance(observed, dict):
        observed_dict = cast(Dict[Any, Any], observed)
        return cast(T, {k: deep_to_raw(v) for k, v in observed_dict.items()})
    elif isinstance(observed, list):
        observed_list = cast(Iterable[Any], observed)
        return cast(T, [deep_to_raw(v) for v in observed_list])
    elif isinstance(observed, tuple):
        observed_tuple = cast(Iterable[Any], observed)
        return cast(T, tuple(deep_to_raw(v) for v in observed_tuple))
    elif isinstance(observed, set):
        observed_set = cast(Iterable[Any], observed)
        return cast(T, {deep_to_raw(v) for v in observed_set})
    elif isinstance(observed, frozenset):
        observed_frozenset = cast(Iterable[Any], observed)
        return cast(T, frozenset(deep_to_raw(v) for v in observed_frozenset))
    return to_raw(observed)


def record_new_reactive_obj(original: T, observed: T) -> T:
    __global_reactive_object_map[id(original)] = observed
    __global_original_object_map[id(observed)] = original
    if DEV:
        print(
            f'record_new_reactive_obj: {original} at {hex(id(original))} ({id(original)}) -> {observed} at {hex(id(observed))} ({id(observed)})'
        )
    return observed


def mark_raw(instance: T) -> T:
    if DEV:
        print(f'mark_raw: {instance} at {hex(id(instance))} ({id(instance)})')
    __marked_raw_set.add(id(instance))
    return instance


def is_marked_raw(instance: object) -> bool:
    return id(instance) in __marked_raw_set



REF_VALUE = '__REF_VALUE__'
FLAG_OF_REF = '__IS_REF__'
FLAG_OF_COMPUTED_REF = '__IS_COMPUTED_REF__'

REACTIVITY_VALUE = '__REACTIVITY_VALUE__'
FLAG_OF_REACTIVE = '__IS_REACTIVE__'

FLAG_OF_SKIP = '__REACTIVE_SKIP__'

FLAG_OF_READONLY = '__IS_READONLY__'



# reactivity/effect.py

from typing import Any, Callable, List, Set, TypeVar, Union

from .definations import ReactiveEffectDef
from .utils import cleanup_effect
from .vars import active_effect_stack

T = TypeVar('T')


class ReactiveEffect(ReactiveEffectDef[T]):
    active: bool
    fn: Callable[[], T]
    scheduler: Union[Callable[[], None], None]
    computed: Union[Any, None]  # type: ComputedRefImpl[T]
    deps: 'List[Set[ReactiveEffectDef[Any]]]'

    def __init__(self, fn: Callable[[], T], scheduler: Union[Callable[[], None], None] = None) -> None:
        self.active = True
        self.fn = fn
        self.scheduler = scheduler
        self.computed = None
        self.deps = []

    def run(self) -> T:
        if not self.active:
            return self.fn()
        try:
            active_effect_stack.append(self)
            return self.fn()
        finally:
            if active_effect_stack:
                active_effect_stack.pop()

    def stop(self) -> None:
        cleanup_effect(self)
        self.active = False

    def __call__(self) -> T:
        return self.run()


def effect(update: Callable[[], T]) -> ReactiveEffect[T]:

    def wrapper():
        e = ReactiveEffect(update)
        e.run()
        return e

    return wrapper()


__all__ = ['effect', 'ReactiveEffect']



from typing import Any, Set

from .definations import ReactiveEffectDef
from .vars import active_effect_stack


def cleanup_effect(effect: ReactiveEffectDef[Any]) -> None:
    if not effect.deps:
        return
    for dep in effect.deps:
        dep.remove(effect)
    effect.deps.clear()


def track_effects(dep: Set[ReactiveEffectDef[Any]]):
    active_effect = active_effect_stack[-1]
    dep.add(active_effect)
    if dep not in active_effect.deps:
        active_effect.deps.append(dep)


def trigger_effects(effects: Set[ReactiveEffectDef[Any]]):
    effect_list = list(effects)
    for effect in effect_list:
        if effect.computed is not None:
            trigger_effect(effect)
    for effect in effect_list:
        if effect.computed is None:
            trigger_effect(effect)


def trigger_effect(effect: ReactiveEffectDef[Any]) -> None:
    if effect.scheduler is None:
        effect.run()
    else:
        effect.scheduler()



# pyright: reportMissingTypeStubs=false

from inspect import isfunction, signature
from typing import (Any, Callable, Dict, List, Sequence, TypeVar, Union, cast, overload)

from reactivity.effect import effect
from reactivity.reactive.utils import is_reactive
from reactivity.reactive.vars import immutable_builtin_types
from reactivity.ref.definitions import Ref
from reactivity.ref.utils import is_ref

T = TypeVar('T')

StopHandle = Callable[[], None]

WatchSource = Union[Ref[T], Callable[[], T], T]

OnCleanup = Callable[[Callable[[], None]], None]

WatchCallback = Union[Callable[[T, T, OnCleanup], Any], Callable[[T, T], Any], Callable[[T], Any], Callable[[], Any]]


def __get_fn_params_count(fn: Callable[..., Any]) -> int:
    return len(signature(fn).parameters)


@overload
def watch_effect(update: Callable[[], Any]) -> StopHandle:
    ...


@overload
def watch_effect(update: Callable[[OnCleanup], Any]) -> StopHandle:
    ...


def watch_effect(update: Union[Callable[[], Any], Callable[[OnCleanup], None]], **kwargs: Any) -> StopHandle:
    default_kwargs: Dict[Any, Any] = {}
    for k in kwargs:
        if k not in default_kwargs:
            raise TypeError(f'Unknown keyword argument {k} for watch_effect().')
        if type(kwargs[k]) != type(default_kwargs[k]):
            raise TypeError(
                f'The type of {k} must be {type(default_kwargs[k])} for watch_effect(), but got {type(kwargs[k])}.')

    stop_flag = False

    update_fn_params_count = __get_fn_params_count(update)

    cleanup_callback: Union[Callable[[], None], None] = None

    def stop():
        nonlocal stop_flag, cleanup_callback

        stop_flag = True

        if cleanup_callback is not None:
            cleanup_callback()
            cleanup_callback = None

    def watch_effect_wrapper():
        nonlocal stop_flag

        if stop_flag:
            return

        if cleanup_callback is not None:
            cleanup_callback()

        def cleanup(cb: Callable[[], None]):
            nonlocal cleanup_callback
            cleanup_callback = cb

        if update_fn_params_count == 0:
            update_fn = cast(Callable[[], None], update)
            update_fn()
        else:
            update_fn = cast(Callable[[OnCleanup], None], update)
            update_fn(cleanup)

    effect(watch_effect_wrapper)

    return stop


def __deeply_walk(obj: object) -> None:
    if is_ref(obj):
        obj = cast(Ref[Any], obj).value
    if not is_reactive(obj):
        return
    if isinstance(obj, dict):
        obj = cast(Dict[Any, Any], obj)
        for k in obj:
            __deeply_walk(obj[k])
    elif isinstance(obj, (list, set)):
        obj = cast(Sequence[Any], obj)
        for v in obj:
            __deeply_walk(v)


@overload
def watch(
    source: WatchSource[T],
    callback: WatchCallback[T],
    *,
    deep: bool = False,
    immediate: bool = False,
    **kwargs: Any,
) -> StopHandle:
    ...


@overload
def watch(source: Sequence[WatchSource[T]],
          callback: WatchCallback[List[T]],
          *,
          deep: bool = False,
          immediate: bool = False,
          **kwargs: Dict[Any, Any]) -> StopHandle:
    ...


def watch(source: Union[WatchSource[T], Sequence[WatchSource[T]]],
          callback: Union[WatchCallback[T], WatchCallback[List[T]]],
          *,
          deep: bool = False,
          immediate: bool = False,
          **kwargs: Any) -> StopHandle:
    '''
    Watch a source and run a callback when the source changes.
    
    Args:
        source: The source to watch.
        callback: The callback to run when the source changes.

    Keyword Args:
        deep: Whether to watch the source deeply. Defaults to False.
        immediate: Whether to run the callback immediately. Defaults to False.
    
    Returns:
        A function to stop watching.
    '''
    default_kwargs = {
        'immediate': False,
        'deep': False,
    }
    kwargs.update({
        'immediate': immediate,
        'deep': deep,
    })
    for k in kwargs:
        if k not in default_kwargs:
            raise TypeError(f'Unknown keyword argument {k} for watch().')
        if type(kwargs[k]) != type(default_kwargs[k]):
            raise TypeError(
                f'The type of {k} must be {type(default_kwargs[k])} for watch(), but got {type(kwargs[k])}.')
    immediate = kwargs.get('immediate', default_kwargs['immediate'])
    deep = kwargs.get('deep', default_kwargs['deep'])

    stop_flag = False

    cleanup_callback: Union[Callable[[], None], None] = None

    def stop():
        nonlocal stop_flag, cleanup_callback

        stop_flag = True

        if cleanup_callback is not None:
            cleanup_callback()
            cleanup_callback = None

    if isinstance(source, Sequence):
        source = cast(Sequence[WatchSource[T]], source)
        if is_reactive(source):
            src_list = cast(List[WatchSource[T]], [source])
            single_src_mode = True
        else:
            src_list = list(source)
            single_src_mode = False
    else:
        src_list = cast(List[WatchSource[T]], [source])
        single_src_mode = True
    for src in src_list:
        if not is_reactive(src) and not is_ref(src) and not isfunction(src):
            raise TypeError(f'Invalid watch source type: {type(src)} for watch().')

    N = len(src_list)
    old_value_list: List[Union[T, None]] = [None] * N
    is_first_run = True

    def gen_fn(i: int) -> Callable[[], T]:
        src = src_list[i]
        need_walk_deep = deep
        if is_reactive(src):
            need_walk_deep = True

        def fn() -> T:
            if isfunction(src):
                res = src()
            elif is_ref(src):
                res = cast(Ref[T], src).value
            elif is_reactive(src):
                res = src
            else:
                raise TypeError(f'Invalid watch source type: {type(src)} for watch().')
            res = cast(T, res)
            if need_walk_deep:
                __deeply_walk(res)
            return res

        return fn

    fn_list = [gen_fn(i) for i in range(N)]
    params_cnt = __get_fn_params_count(callback)

    def calc_new_value_list() -> List[T]:
        unfilled_new_value_list: List[Union[T, None]] = [None] * N
        for i in range(N):
            fn = fn_list[i]
            new_value = fn()
            unfilled_new_value_list[i] = new_value
        new_value_list = cast(List[T], unfilled_new_value_list)
        return new_value_list

    def watch_wrapper():
        nonlocal stop_flag, is_first_run

        if stop_flag:
            return

        if cleanup_callback is not None:
            cleanup_callback()

        def cleanup(cb: Callable[[], None]):
            nonlocal cleanup_callback
            cleanup_callback = cb

        try:
            new_value_list = calc_new_value_list()

            skip = True
            # Check if the value has changed, if not, skip the callback.
            for i in range(N):
                new_value = new_value_list[i]
                old_value = old_value_list[i]
                if new_value != old_value:
                    skip = False
                    break
                # If the new_value and the old_value are the same object, we cannot judge if the value has changed.
                # So we still need to run the callback.
                if not isinstance(new_value, immutable_builtin_types) and new_value is old_value:
                    skip = False
                    break
            # Check if it's the first run and `immediate` is False, if so, skip the callback.
            if is_first_run and not immediate:
                old_value_list[:] = new_value_list
                skip = True
            if skip:
                return

            if params_cnt == 0:
                cb = cast(Callable[[], None], callback)
                cb()
            elif params_cnt == 1:
                if single_src_mode:
                    cb = cast(Callable[[T], None], callback)
                    cb(new_value_list[0])
                else:
                    cb = cast(Callable[[List[T]], None], callback)
                    cb(new_value_list)
            elif params_cnt == 2:
                if single_src_mode:
                    cb = cast(Callable[[T, Union[T, None]], None], callback)
                    cb(new_value_list[0], old_value_list[0])
                else:
                    cb = cast(Callable[[List[T], List[Union[T, None]]], None], callback)
                    cb(new_value_list, old_value_list[:])
            elif params_cnt == 3:
                if single_src_mode:
                    cb = cast(Callable[[T, Union[T, None], OnCleanup], None], callback)
                    cb(new_value_list[0], old_value_list[0], cleanup)
                else:
                    cb = cast(Callable[[List[T], List[Union[T, None]], OnCleanup], None], callback)
                    cb(new_value_list, old_value_list[:], cleanup)
            else:
                raise TypeError(
                    f'Invalid callback function for watch(). The callback function must have 0, 1 or 2 parameters, but got {params_cnt} parameters.'
                )

            old_value_list[:] = new_value_list
        finally:
            if is_first_run:
                is_first_run = False

    effect(watch_wrapper)

    return stop


__all__ = ['watch', 'watch_effect']
